import json
from langchain.callbacks import get_openai_callback
from repo_chat import chain_manager
import time


class RetrievalChain:
    """
    A class for chatting with a large language model using document retrieval from a GitHub repo.
    It leverages the langchain library to facilitate the retrieval and chat operations.
    """

    def __init__(self, vectorstore, repo):
        """
        Initialize the class with the given vectorstore
        """
        self.vectorstore = vectorstore
        self.repo = repo
        self.chainlogs = []

    def log_entry(self, chain_name, input_data, output_data, exec_time):
        """Log the given method, input_data, output_data and execution time"""
        self.chainlogs.append({
            "chain_name": chain_name,
            "input_data": json.dumps(input_data),
            "output_data": output_data['text'],
            "execution_time": exec_time
        })

    def get_chain_inputs(self, query, get_docs=True):
        """
        Retrieve similar documents from vectorstore based on the given query.
        Prepare the chain inputs using these documents.
        """
        chain_inputs = {
            "query": query,
            "repo": self.repo,
        }
        if get_docs:
            docs = self.vectorstore.similarity_search_with_score(query, k=5)
            chain_inputs["similar_documents"] = [doc.page_content for doc, _ in docs]
        return chain_inputs
    
    def call_chain(self, chain, input_data):
        """Call the given chain and log the execution time"""
        start_time = time.time()
        output = chain(input_data)
        exec_time = time.time() - start_time
        self.log_entry(chain.name, input_data, output, exec_time)
        return output

    def process_query(self, query, context_validator, run_query, context_threshold=60):
        """
        Process the given query using the context_validator and run_query chains.
        If sufficient context is present, return the result of run_query.
        """
        chain_inputs = self.get_chain_inputs(query)
        val_resp = self.call_chain(context_validator, chain_inputs)
        sufficient_context = int(val_resp["text"].lower())
        if sufficient_context > context_threshold:
            return self.call_chain(run_query, chain_inputs)
        else:
            return None

    def iterate_through_queries(
        self, upgrade_query, query, context_validator, run_query, context_threshold
    ):
        """
        For each query retrieved from the upgrade_query response,
        process the query and return the answer if found.
        """
        chain_inputs = self.get_chain_inputs(query, get_docs=False)
        upgrade_query_resp = self.call_chain(upgrade_query, chain_inputs)
        for q in upgrade_query_resp["text"].split("\n"):
            refined_query = q.lstrip("0123456789. ")
            answer = self.process_query(
                refined_query, context_validator, run_query, context_threshold
            )
            if answer is not None:
                return answer
        return None

    def manage_workflow(self, query, context_validator, run_query):
        """Manage workflow for upgrading the query and processing it"""
        context_threshold = 60
        upgrade_query = chain_manager.get_chain("UPGRADE_QUERY")

        while True:
            answer = self.iterate_through_queries(
                upgrade_query, query, context_validator, run_query, context_threshold
            )
            if answer is not None or context_threshold <= 0:
                break
            context_threshold -= 10
        return answer

    def chat(self, query):
        """
        Begin chat with the QA chain
        Initialize context_validator and run_query chains.
        """
        context_validator = chain_manager.get_chain("CONTEXT_VALIDATOR")
        run_query = chain_manager.get_chain("RUN_QUERY_RAG")

        with get_openai_callback() as cb:
            answer = self.process_query(query, context_validator, run_query)
            if answer is not None:
                return answer
            else:
                return self.manage_workflow(query, context_validator, run_query)


class RawChain:
    """ Chain specific for raw code input """
    def __init__(self, repo_data):
        """ Initialize the basic chain object """
        self.repo_data = repo_data

    def chat(self, query):
        """ Chat with the basic chain object """
        raw_chain = chain_manager.get_chain("RAW_CODE")

        with get_openai_callback() as cb:
            chain_inputs = {
                "query": query,
                "repo_data": self.repo_data,
            }
            output = raw_chain(chain_inputs)
            return output



import concurrent.futures
import pandas as pd
from repo_chat.chat_utils import  RetrievalChain
from repo_chat import chain_manager

class CriticChain:
    def __init__(self):
        self.chain = chain_manager.get_chain("CRITIC")

    def score(self, query, response, repo):
        """Score a given response to a query"""
        inputs = {"query": query, "response": response, "repo": repo}
        return self.chain(inputs)


class QueryEvaluator:
    def __init__(self, get_vectorstore, query, repo, runs_per_query=5):
        self.get_vectorstore = get_vectorstore
        self.query = query
        self.runs_per_query = runs_per_query
        self.repo = repo
        self.critic = CriticChain()

    def score_response(self, chain, query, query_id):
        """Score a response using CriticChain"""
        response = chain.chat(query)
        score = self.critic.score(query, response["text"], self.repo)
        response['score'] = score['text']

        # add query_id to response and chainlogs
        response['query_id'] = query_id
        for log in chain.chainlogs:
            log['query_id'] = query_id

        return {
            'chainlogs': chain.chainlogs,
            'response': response
        }

    def evaluate(self):
        """Evaluate RetrievalChain using CriticChain to score responses"""
        responses = []
        for query_id in range(self.runs_per_query):
            vectorstore = self.get_vectorstore()
            chain = RetrievalChain(vectorstore, self.repo)
            response = self.score_response(chain, self.query, query_id)
            responses.append(response)

        return responses


class MultiQueryEvaluator:
    def __init__(self, get_vectorstore, queries, repo, runs_per_query=5):
        self.get_vectorstore = get_vectorstore
        self.queries = queries
        self.runs_per_query = runs_per_query
        self.repo = repo
        self.log_user_message()

    def log_user_message(self):
        print("MultiQueryEvaluator initialized with the following configurations:")
        print(f"Number of queries: {len(self.queries)}")
        print(f"Runs per query: {self.runs_per_query}")
        print(f"Total evaluations: {len(self.queries) * self.runs_per_query}")
        print("Parallelized by query using concurrent.futures.ProcessPoolExecutor")

    def run_query(self, query):
        """Run QueryEvaluator for a given query"""
        response = QueryEvaluator(
            self.get_vectorstore, query, self.repo, self.runs_per_query,
        ).evaluate()
        return query, response

    def evaluate(self, max_workers=4):
        """Evaluate multiple queries in parallel"""
        responses = {}
        with concurrent.futures.ProcessPoolExecutor(max_workers=max_workers) as executor:
            for query, response in executor.map(self.run_query, self.queries):
                responses[query] = response
        self.responses = responses

    def flatten_responses(self):
        """
        Convert responses into two dataframes:
        1. chainlogs_df: chainlogs for each query-response pair
        2. response_df: data from each query-response pair
        """
        chainlogs_dfs = []
        response_dfs = []

        for key, data in self.responses.items():
            for obj in data:

                df_chainlogs = pd.json_normalize(obj['chainlogs'])
                df_chainlogs['query'] = key
                chainlogs_dfs.append(df_chainlogs)

                df_response = pd.DataFrame([obj['response']])
                df_response['similar_documents'] = df_response['similar_documents'].apply(lambda x: ','.join(x))
                df_response['query'] = key
                response_dfs.append(df_response)

        chainlogs_df = pd.concat(chainlogs_dfs, ignore_index=True)
        response_df = pd.concat(response_dfs, ignore_index=True)

        return chainlogs_df, response_df





from typing import Any
import dotenv
import os
from langchain.prompts import PromptTemplate
from langchain.chat_models import ChatOpenAI
from langchain.chains import LLMChain
from repo_chat import templates

dotenv.load_dotenv()
OPENAI_API_KEY = os.getenv("OPENAI_API_KEY")


class CustomChain:
    """
    A class to manage creating and storing chains/openai credentials.
    """

    def __init__(self, name, temperature=0.1, model_name = "gpt-3.5-turbo"):
        """
        Initialize the class by loading the OPENAI_API_KEY from the .env file.
        """
        self.name = name
        self.model_name = model_name
        self.temperature = temperature
        self.openai_api_key = OPENAI_API_KEY
        self.chain = self.create_chain(**getattr(templates, name))

    def create_chain(self, input_variables, template):
        """Create an LLMChain given model, input_variables, template and temperature."""
        prompt = PromptTemplate(input_variables=input_variables, template=template)
        llm = ChatOpenAI(
            openai_api_key=self.openai_api_key,
            model_name=self.model_name,
            temperature=self.temperature,
        )
        return LLMChain(llm=llm, prompt=prompt)

    def __call__(self, input_data: Any):
        """Call the chain with the given input_data."""
        return self.chain(input_data)
    

def get_chain(template_type):
    """Get chain by passing the respective template type."""
    return CustomChain(template_type)



UPGRADE_QUERY = {
    "input_variables": ["query", "repo"],
    "template": """In the context of some github repository ({repo}), a user has submitted the following query:
    {query}

    Based on this query, documents will be retrieved from the repository to provide an answer. However, similarity between the query and the documents is not the only important factor - the retrieved documents must also be relevant and help answer the user's query effectively.
    Revise and expand the query to improve the quality and relevance of the documents retrieved. Consider the user's intent, possible context, and the nature of the code repository. If appropriate, feel free to rephrase the query or break down the question into several related, more specific queries. Minimize Tokens. Give me 5 unique revised queries following these guidlines.
    """
}

RUN_QUERY_RAG = {
    "input_variables": ["query", "similar_documents", "repo"],
    "template": """You are an expert software engineering assistant. In the context of some github repository ({repo}), a user has submitted the following query:
    {query}

    The following documents have been retrieved from this repo because they contain information potentially relevant to the query:
    {similar_documents}

    Given your understanding of the query and the information contained within these documents, provide the most accurate and relevant response possible. You are a very knowledgeable expert on the topic so feel free to infer information that is not explicitly stated in the documents. Be super concise! Your response must be in .md format. Minimize Tokens by using paraphrasing.
    """
}

CONTEXT_VALIDATOR = {
    "input_variables": ["query", "similar_documents", "repo"],
    "template": """You are an expert software engineering assistant. In the context of some github repository ({repo}), a user has submitted the following query:
    {query}

    The following documents have been retrieved from this repo because they contain information potentially relevant to the query:
    {similar_documents}

    How sufficient is the provided context for answering the user's query? Please respond with a number 0 (worst - the documents do not provide any information useful for answering the question) to 100 (best - the documents provide all information required for answering the question). Nothing else.
    """
}

CRITIC = {
    "input_variables": ["query", "response", "repo"],
    "template": """In the context of some github repository ({repo}), a user has submitted the following query:
    {query}

    The following response has been generated:
    {response}

    Does the resonse answer the users query about some specific code repository? Score this response on a scale from 0 (completely irrelevant or incorrect) to 100 (perfectly helpful based on the users query). Return the numeric score only with no words.
    """
}

RAW_CODE = {
    "input_variables": ["query", "repo_data"],
    "template": """Here is my code repository:
    {repo_data}

    Regarding the code above:
    {query}
    """
}


from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="repo_chat",
    version="0.1.2",
    author="Jamie Voynow",
    author_email="voynow99@gmail.com",
    description="Using LLMS to facilitate document retrieval and chat operations",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/voynow/repo-chat",
    packages=find_packages(),
    install_requires=[
        'langchain',
        'pandas',
        'matplotlib',
        'numpy',
        'ipywidgets',
        'uuid',
        'python-dotenv',
        'git2vec'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)
